# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os
import subprocess


def git_hook_format_changes(changes):
    """Format Git changes in the way Git hooks expect them in `stdin`.

    :param changes: a dict with keys are Git refs, and values are pairs
                    (old_sha, new_sha)
    :return: formatted changes ready to pipe to a Git hook
    """
    return '\n'.join(
        ' '.join((old_gitsha, new_gitsha, ref))
        for ref, (old_gitsha, new_gitsha) in changes.items())


SKIP = object()


class Hook(object):
    """Allow to call a GitLab hook.

    In Heptapod, we disable by default some of the Git hooks, in order to get
    finer control of them by calling them directly.

    This allows the Mercurial process to send accurate user information in
    its inner pushes to Git repos, as well as to notify GitLab of changes
    at the right stage of Mercurial transaction end.
    """

    def __init__(self, name, repo):
        self.name = name
        self.repo = repo
        self.git_fs_path = self.repo.root[:-3] + '.git'
        shell = repo.ui.config(b'heptapod', b'gitlab-shell')
        if not shell:
            raise RuntimeError("Path to GitLab Shell is unknown")
        self.gitlab_shell = shell

    def __str__(self):
        return "GitLab %r hook wrapper" % self.name

    def environ(self):
        base = self.repo.ui.environ
        if base.get('HEPTAPOD_SKIP_ALL_GITLAB_HOOKS', '').strip() == 'yes':
            return SKIP
        if base.get('HEPTAPOD_SKIP_GITLAB_HOOK', '').strip() == self.name:
            return SKIP

        gl_id = base.get('HEPTAPOD_USERINFO_ID')
        project_id = base.get('HEPTAPOD_PROJECT_ID')

        if gl_id is None:
            raise ValueError("User id not available")
        if project_id is None:
            raise ValueError("Project id not available")

        env = dict(os.environ)
        env['HG_GIT_SYNC'] = 'yes'
        env['GL_ID'] = 'user-' + gl_id
        # Protocol is only to deny access based on blocked protocol,
        # but 'web' is always accepted
        # (see gitlab_access.rb and protocol_access.rb
        # from gitlab-rails/lib/gitlab lib/gitlab)
        env['GL_PROTOCOL'] = 'web'
        env['GL_REPOSITORY'] = 'project-' + project_id
        return env

    def __call__(self, changes):
        """Call GitLab Hook for the given changes.

        :return: numeric return code, output and error.
                 The return code is like for a process does: 0 for success,
                 anything else for failures.
        """
        if not changes:
            self.repo.ui.debug(
                "%s: empty set of changes - nothing to do." % self)
            return 0, '', ''

        try:
            env = self.environ()
        except ValueError as exc:
            msg = "%s, not sending notifications!" % exc.args[0]
            self.repo.ui.warn("%s: %s" % (self, msg))
            return 1, '', msg

        if env is SKIP:
            self.repo.ui.note('%s: bailing (explicitely told not to send)')
            return 0, '', ''

        self.repo.ui.note("%s: sending changes %r" % (self, changes))
        pr = subprocess.Popen(
            [os.path.join(self.gitlab_shell, 'hooks', self.name)],
            env=env,
            cwd=self.git_fs_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE)
        out, err = pr.communicate(input=git_hook_format_changes(changes))
        return pr.returncode, out, err



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later


def print_heptapod_env(repo, *args, **kwargs):
    repo.ui.status(repr(sorted(
        (k, v) for (k, v) in repo.ui.environ.items()
        if k.startswith('HEPTAPOD_'))))
    return 0



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest
from heptapod.testhelpers import (
    LocalRepoWrapper,
    )
from .. import (
    git_hook_format_changes,
    Hook,
    )

ZERO_SHA = "0" * 20


def test_git_hook_format_changes():
    changes = {'branch/newbr': (ZERO_SHA, "beef1234dead" + "a" * 8),
               'topic/default/zz': ("23cafe45" + "0" * 12, "1234" + "b" * 16),
               }
    assert set(git_hook_format_changes(changes).split('\n')) == {
        '00000000000000000000 beef1234deadaaaaaaaa branch/newbr',
        '23cafe45000000000000 1234bbbbbbbbbbbbbbbb topic/default/zz'
    }


def make_repo(base_dir, name):
    wrapper = LocalRepoWrapper.init(
        base_dir.join(name),
        config=dict(heptapod={'gitlab-shell': str(base_dir)}),
        )
    # by default, ui.environ IS os.environ (shared) but that isn't true
    # in WSGI contexts. In that case, it is a plain dict, the WSGI environment
    # indeed. For these tests, we need them to be independent.
    wrapper.repo.ui.environ = {}
    return wrapper


def test_hook(tmpdir):
    wrapper = make_repo(tmpdir, 'proj.hg')
    repo = wrapper.repo
    hook = Hook('post-something', repo)

    repo.ui.environ.update(
        HEPTAPOD_USERINFO_ID='23',
        HEPTAPOD_PROJECT_ID='1024',
    )

    assert str(hook) == "GitLab 'post-something' hook wrapper"
    env = hook.environ()

    assert env.get('HG_GIT_SYNC') == 'yes'
    assert env.get('GL_ID') == 'user-23'
    assert env.get('GL_REPOSITORY') == 'project-1024'

    # we didn't pollute original environment
    assert 'GL_ID' not in repo.ui.environ

    # calling with empty changes
    assert hook(()) == (0, '', '')

    # now with an executable
    # we need a pseudo Git repo
    tmpdir.join('proj.git').ensure(dir=True)

    hook_bin = tmpdir.ensure('hooks', dir=True).join('post-something')
    hook_bin.write('\n'.join((
        "#!/usr/bin/env python",
        "import sys",
        "sys.stderr.write('stderr of the post something hook')",
        "for line in sys.stdin.readlines():",
        "    sys.stdout.write(line)",
        "sys.exit(3)",
    )))
    hook_bin.chmod(0o700)
    changes = {'branch/newbr': (ZERO_SHA, "beef1234dead" + "a" * 8),
               'topic/default/zz': ("23cafe45" + "0" * 12, "1234" + "b" * 16),
               }

    code, out, err = hook(changes)
    assert code, err == (3, 'stderr of the post something hoko')
    assert set(out.splitlines()) == {
        '00000000000000000000 beef1234deadaaaaaaaa branch/newbr',
        '23cafe45000000000000 1234bbbbbbbbbbbbbbbb topic/default/zz'
    }


def test_missing_gitlab_shell(tmpdir):
    repo = LocalRepoWrapper.init(tmpdir).repo
    with pytest.raises(RuntimeError) as exc_info:
        Hook('post-something', repo)
    assert "Path to GitLab Shell" in exc_info.value.args[0]


def test_hook_missing_user(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    hook = Hook('post-something', repo)

    repo.ui.environ['HEPTAPOD_PROJECT_ID'] = '1024'
    with pytest.raises(ValueError) as exc_info:
        hook.environ()
    assert 'User id' in exc_info.value.args[0]

    # and that gets catched, does not break the process
    code, out, err = hook(dict(master=(0, 1)))
    assert code != 0


def test_hook_missing_project(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    hook = Hook('post-something', repo)

    repo.ui.environ['HEPTAPOD_USERINFO_ID'] = '3'
    with pytest.raises(ValueError) as exc_info:
        hook.environ()
    assert 'Project id' in exc_info.value.args[0]

    # and that gets catched, does not break the process
    code, out, err = hook(dict(master=(0, 1)))
    assert code != 0


def test_hook_skip(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    repo.ui.environ['HEPTAPOD_SKIP_GITLAB_HOOK'] = 'pre-rcv'
    hook = Hook('pre-rcv', repo)
    assert hook(dict(master=(0, 1))) == (0, '', '')


def test_hook_skip_all(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    repo.ui.environ['HEPTAPOD_SKIP_ALL_GITLAB_HOOKS'] = 'no'
    hook = Hook('pre-rcv', repo)
    # it's been called and failed because of lack of User id
    assert hook(dict(master=(0, 1)))[0] == 1

    repo.ui.environ['HEPTAPOD_SKIP_ALL_GITLAB_HOOKS'] = 'yes'
    hook = Hook('pre-rcv', repo)
    assert hook(dict(master=(0, 1))) == (0, '', '')
